# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Benchmark API for Intel Neural Compressor."""

import argparse
import os
import re
import subprocess
import sys

import psutil

from neural_compressor.common.utils import Statistics, get_workspace, logger

description = """
##################################################################################################################
This is the command used to launch the Intel CPU performance benchmark, supports both Linux and Windows platform.
To get the peak performance on Intel Xeon CPU, we should avoid crossing NUMA node in one instance.
By default, `incbench` will trigger 1 instance on the first NUMA node.

Params in `incbench`:
 - num_instances            Default to 1.
 - num_cores_per_instance   Default to None.
 - C, cores                 Default to 0-${num_cores_on_NUMA-1}, decides the visible core range.
 - cross_memory             Default to False, decides whether to allocate memory cross NUMA.
                                Note: Use it only when memory for instance is not enough.

# General use cases:
1. `incbench main.py`: run 1 instance on NUMA:0.
2. `incbench --num_i 2 main.py`: run 2 instances on NUMA:0.
3. `incbench --num_c 2 main.py`: run multi-instances with 2 cores per instance on NUMA:0.
4. `incbench -C 24-47 main.py`: run 1 instance on COREs:24-47.
5. `incbench -C 24-47 --num_c 4 main.py`: run multi-instances with 4 COREs per instance on COREs:24-47.

Note:
    - `num_i` works the same as `num_instances`
    - `num_c` works the same as `num_cores_per_instance`
##################################################################################################################
"""


def get_linux_numa_info():
    """Collect numa/socket information on linux system.

    Returns:
        numa_info (dict):   demo: {numa_index: {"physical_cpus": "xxx"; "logical_cpus": "xxx"}}
                            E.g.    numa_info = {
                                        0: {"physical_cpus": "0-23", "logical_cpus": "0-23,48-71"},
                                        1: {"physical_cpus": "24-47", "logical_cpus": "24-47,72-95"}
                                    }
    """
    result = subprocess.run(["lscpu"], capture_output=True, text=True)
    output = result.stdout

    numa_info = {}
    for line in output.splitlines():
        # demo: "NUMA node0 CPU(s): 0-3"
        node_match = re.match(r"^NUMA node(\d+) CPU\(s\):\s+(.*)$", line)
        if node_match:
            node_id = int(node_match.group(1))
            cpus = node_match.group(2).strip()
            numa_info[node_id] = {
                "physical_cpus": cpus.split(",")[0],
                "logical_cpus": ",".join(cpus.split(",")),
            }

    # if numa_info is not collected, we go back to socket_info
    if not numa_info:  # pragma: no cover
        for line in output.splitlines():
            # demo: "Socket(s):             2"
            socket_match = re.match(r"^Socket\(s\):\s+(.*)$", line)
            if socket_match:
                num_socket = int(socket_match.group(1))
        # process big cores (w/ physical cores) and small cores (w/o physical cores)
        physical_cpus = psutil.cpu_count(logical=False)
        logical_cpus = psutil.cpu_count(logical=True)
        physical_cpus_per_socket = physical_cpus // num_socket
        logical_cpus_per_socket = logical_cpus // num_socket
        for i in range(num_socket):
            physical_cpus_str = str(i * physical_cpus_per_socket) + "-" + str((i + 1) * physical_cpus_per_socket - 1)
            if num_socket == 1:
                logical_cpus_str = str(i * logical_cpus_per_socket) + "-" + str((i + 1) * logical_cpus_per_socket - 1)
            else:
                remain_cpus = logical_cpus_per_socket - physical_cpus_per_socket
                logical_cpus_str = (
                    physical_cpus_str
                    + ","
                    + str(i * (remain_cpus) + physical_cpus)
                    + "-"
                    + str((i + 1) * remain_cpus + physical_cpus - 1)
                )
            numa_info[i] = {
                "physical_cpus": physical_cpus_str,
                "logical_cpus": logical_cpus_str,
            }
    return numa_info


def get_windows_numa_info():
    """Collect socket information on Windows system due to no available numa info.

    Returns:
        numa_info (dict):   demo: {numa_index: {"physical_cpus": "xxx"; "logical_cpus": "xxx"}}
                            E.g.    numa_info = {
                                        0: {"physical_cpus": "0-23", "logical_cpus": "0-23,48-71"},
                                        1: {"physical_cpus": "24-47", "logical_cpus": "24-47,72-95"}
                                    }
    """
    # pylint: disable=import-error
    # pragma: no cover
    import wmi

    c = wmi.WMI()
    processors = c.Win32_Processor()
    socket_designations = set()
    for processor in processors:
        socket_designations.add(processor.SocketDesignation)
    num_socket = len(socket_designations)
    physical_cpus = sum(processor.NumberOfCores for processor in processors)
    logical_cpus = sum(processor.NumberOfLogicalProcessors for processor in processors)
    physical_cpus_per_socket = physical_cpus // num_socket
    logical_cpus_per_socket = logical_cpus // num_socket

    numa_info = {}
    for i in range(num_socket):
        physical_cpus_str = str(i * physical_cpus_per_socket) + "-" + str((i + 1) * physical_cpus_per_socket - 1)
        if num_socket == 1:
            logical_cpus_str = str(i * logical_cpus_per_socket) + "-" + str((i + 1) * logical_cpus_per_socket - 1)
        else:
            remain_cpus = logical_cpus_per_socket - physical_cpus_per_socket
            logical_cpus_str = (
                physical_cpus_str
                + ","
                + str(i * (remain_cpus) + physical_cpus)
                + "-"
                + str((i + 1) * remain_cpus + physical_cpus - 1)
            )
        numa_info[i] = {
            "physical_cpus": physical_cpus_str,
            "logical_cpus": logical_cpus_str,
        }
    return numa_info


def dump_numa_info():
    """Fetch NUMA info and dump stats in shell, return numa_info.

    Returns:
        numa_info (dict): {numa_node_index: list of Physical CPUs in this numa node, ...}
    """
    if psutil.WINDOWS:  # pragma: no cover
        numa_info = get_windows_numa_info()
    elif psutil.LINUX:
        numa_info = get_linux_numa_info()
    else:  # pragma: no cover
        logger.error(f"Unsupported platform detected: {sys.platform}, only supported on Linux and Windows")

    # dump stats to shell
    field_names = ["NUMA node", "Physical CPUs", "Logical CPUs"]
    output_data = []
    for op_type in numa_info.keys():
        field_results = [op_type, numa_info[op_type]["physical_cpus"], numa_info[op_type]["logical_cpus"]]
        output_data.append(field_results)
    Statistics(output_data, header="CPU Information", field_names=field_names).print_stat()

    # parse numa_info for ease-of-use
    for n in numa_info:
        numa_info[n] = parse_str2list(numa_info[n]["physical_cpus"])
    return numa_info


def parse_str2list(cpu_ranges):
    """Parse '0-4,7,8' into [0,1,2,3,4,7,8] for machine readable."""
    cpus = []
    ranges = cpu_ranges.split(",")
    for r in ranges:
        if "-" in r:
            try:
                start, end = r.split("-")
                cpus.extend(range(int(start), int(end) + 1))
            except ValueError:  # pragma: no cover
                raise ValueError(f"Invalid range: {r}")
        else:
            try:
                cpus.append(int(r))
            except ValueError:  # pragma: no cover
                raise ValueError(f"Invalid number: {r}")
    return cpus


def format_list2str(cpus):
    """Format [0,1,2,3,4,7,8] back to '0-4,7,8' for human readable."""
    if not cpus:  # pragma: no cover
        return ""
    cpus = sorted(set(cpus))
    ranges = []
    start = cpus[0]
    end = start
    for i in range(1, len(cpus)):
        if cpus[i] == end + 1:
            end = cpus[i]
        else:
            if start == end:
                ranges.append(f"{start}")
            else:
                ranges.append(f"{start}-{end}")
            start = cpus[i]
            end = start
    if start == end:
        ranges.append(f"{start}")
    else:
        ranges.append(f"{start}-{end}")
    return ",".join(ranges)


def get_reversed_numa_info(numa_info):
    """Reverse numa_info."""
    reversed_numa_info = {}
    for n, cpu_info in numa_info.items():
        for i in cpu_info:
            reversed_numa_info[i] = n
    return reversed_numa_info


def get_numa_node(core_list, reversed_numa_info):
    """Return numa node used in current core_list."""
    numa_set = set()
    for c in core_list:
        assert c in reversed_numa_info, "Cores should be in physical CPUs"
        numa_set.add(reversed_numa_info[c])
    return numa_set


def set_cores_for_instance(args, numa_info):
    """Set cores for each instance based on the input args.

    All use cases are listed below:
        Params: a=num_instance; b=num_cores_per_instance; c=cores;
            - no a, b, c: a=1, c=numa:0
            - no a, b: a=1, c=c
            - no a, c: a=numa:0/b, c=numa:0
            - no b, c: a=a, c=numa:0
            - no a: a=numa:0/b, c=c
            - no b: a=a, c=c
            - no c: a=a, c=a*b
            - a, b, c: a=a, c=a*b

    Args:
        args (argparse): arguments for setting different configurations
        numa_info (dict): {numa_node_index: list of Physical CPUs in this numa node, ...}

    Returns:
        core_list_per_instance (dict): {"instance_index": ["node_index", "cpu_index", num_cpu]}
    """
    available_cores_list = []
    for n in numa_info:
        available_cores_list.extend(numa_info[n])
    # preprocess args.cores to set default values
    if args.cores is None:
        if args.num_cores_per_instance and args.num_instances:
            target_cores = args.num_instances * args.num_cores_per_instance
            assert target_cores <= len(available_cores_list), (
                "Invalid configuration: num_instances * num_cores_per_instance = "
                + "{} exceeds the number of physical CPUs = {}.".format(target_cores, len(available_cores_list))
            )
            cores_list = list(range(target_cores))
            # log for cores in use
            logger.info("num_instances * num_cores_per_instance = {} cores are used.".format(target_cores))
        else:
            # default behavior, only use numa:0
            cores_list = numa_info[0]
            # log for cores in use
            logger.info("By default, Intel Neural Compressor uses all cores on numa:0.")
    else:
        cores_list = parse_str2list(args.cores)
        # log for cores available
        logger.info("{} cores are available.".format(len(cores_list)))
        if args.num_cores_per_instance and args.num_instances:
            target_cores = args.num_instances * args.num_cores_per_instance
            assert target_cores <= len(cores_list), (
                "Invalid configuration: num_instances * num_cores_per_instance = "
                + "{} exceeds the number of available CPUs = {}.".format(target_cores, len(cores_list))
            )
            cores_list = cores_list[:target_cores]

    # preprocess args.num_instances to set default values
    if args.num_instances is None:
        if args.num_cores_per_instance:
            assert args.num_cores_per_instance <= len(cores_list), (
                "Invalid configuration: num_cores_per_instance = "
                + "{} exceeds the number of available CPUs = {}.".format(args.num_cores_per_instance, len(cores_list))
            )
            args.num_instances = len(cores_list) // args.num_cores_per_instance
            target_cores = args.num_instances * args.num_cores_per_instance
            cores_list = cores_list[:target_cores]
        else:
            args.num_instances = 1
            logger.info("By default, Intel Neural Compressor triggers only one instance.")
    else:
        assert args.num_instances <= len(
            cores_list
        ), "Invalid configuration: num_instances = " + "{} exceeds the number of available CPUs = {}.".format(
            args.num_instances, len(cores_list)
        )

    ### log for instances number and cores in use
    if args.num_instances == 1:
        logger.info("1 instance is triggered.")
    else:
        logger.info("{} instances are triggered.".format(args.num_instances))
    if len(cores_list) == 1:
        logger.info("1 core is in use.")
    else:
        logger.info("{} cores are in use.".format(len(cores_list)))

    # only need to process num_cores_per_instance now
    core_list_per_instance = {}
    # num_cores_per_instance = all_cores / num_instances
    num_cores_per_instance = len(cores_list) // args.num_instances
    for i in range(args.num_instances):
        core_list_per_instance[i] = cores_list[i * num_cores_per_instance : (i + 1) * num_cores_per_instance]
    if len(cores_list) % args.num_instances != 0:  # pragma: no cover
        last_index = args.num_instances - 1
        core_list_per_instance[last_index] = cores_list[last_index * num_cores_per_instance :]

    # convert core_list_per_instance = {"instance_index": cpu_index_list}
    #                                -> {"instance_index": ["node_index", "cpu_index", num_cores]}
    reversed_numa_info = get_reversed_numa_info(numa_info)
    for i, core_list in core_list_per_instance.items():
        core_list_per_instance[i] = [
            format_list2str(get_numa_node(core_list, reversed_numa_info)),
            format_list2str(core_list),
            len(core_list),
        ]

    # dump stats to shell
    field_names = ["Instance", "NUMA node", "Physical CPUs", "Number of cores"]
    output_data = []
    for i, core_list in core_list_per_instance.items():
        field_results = [i + 1, core_list[0], core_list[1], core_list[2]]
        output_data.append(field_results)
    Statistics(output_data, header="Instance Binding Information", field_names=field_names).print_stat()
    return core_list_per_instance


def generate_prefix(args, core_list):
    """Generate the command prefix with `numactl` (Linux) or `start` (Windows) command.

    Args:
        args (argparse): arguments for setting different configurations
        core_list: ["node_index", "cpu_index", num_cpu]

    Returns:
        command_prefix (str): command_prefix with specific core list for Linux or Windows.
    """
    if sys.platform in ["linux"] and os.system("numactl --show >/dev/null 2>&1") == 0:
        if args.cross_memory:
            return "OMP_NUM_THREADS={} numactl -l -C {}".format(core_list[2], core_list[1])
        else:
            return "OMP_NUM_THREADS={} numactl -m {} -C {}".format(core_list[2], core_list[0], core_list[1])
    elif sys.platform in ["win32"]:  # pragma: no cover
        socket_id = core_list[0]
        from functools import reduce

        hex_core = hex(reduce(lambda x, y: x | y, [1 << p for p in parse_str2list(core_list[1])]))
        return "start /B /WAIT /node {} /affinity {}".format(socket_id, hex_core)
    else:  # pragma: no cover
        return ""


def run_multi_instance_command(args, core_list_per_instance, raw_cmd):
    """Build and trigger commands for multi-instances with subprocess.

    Args:
        args (argparse): arguments for setting different configurations
        core_list_per_instance (dict): {"instance_index": ["node_index", "cpu_index", num_cpu]}
        raw_cmd (str): script.py and parameters for this script
    """
    instance_cmd = ""
    if not os.getenv("PYTHON_PATH"):  # pragma: no cover
        logger.info("The interpreter path is not set, using string `python` as command.")
        logger.info("To replace it, use `export PYTHON_PATH=xxx`.")
    interpreter = os.getenv("PYTHON_PATH", "python")
    workspace_dir = get_workspace()
    logfile_process_map = {}
    logfile_dict = {}
    for i, core_list in core_list_per_instance.items():
        # build cmd and log file path
        prefix = generate_prefix(args, core_list)
        instance_cmd = "{} {} {}".format(prefix, interpreter, raw_cmd)
        logger.info(f"Instance {i+1}: {instance_cmd}")
        instance_log_file = "{}_{}_{}C.log".format(i + 1, len(core_list_per_instance), core_list[2])
        instance_log_file = os.path.join(workspace_dir, instance_log_file)
        # trigger subprocess
        p = subprocess.Popen(
            instance_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True
        )  # nosec
        # log_file_path: [process_object, instance_command, instance_index]
        logfile_process_map[instance_log_file] = [p, instance_cmd, i + 1]
        logfile_dict[i + 1] = instance_log_file

    # Dump each instance's standard output to the corresponding log file
    for instance_log_file, p_cmd_i in logfile_process_map.items():
        # p.communicate() reads std to avoid dead-lock, p.wait() only return.
        stdout, stderr = p_cmd_i[0].communicate()  # stderr is merged to stdout, so it's None
        with open(instance_log_file, "w", 1, encoding="utf-8") as log_file:
            log_file.write(f"[COMMAND]: {p_cmd_i[1]}\n")
            log_file.write(stdout.decode())
        logger.info(f"The log of instance {p_cmd_i[2]} is saved to {instance_log_file}")

    return logfile_dict


def summary_latency_throughput(logfile_dict):
    """Get the summary of the benchmark."""
    throughput_pattern = r"[T,t]hroughput:\s*([0-9]*\.?[0-9]+)\s*([a-zA-Z/]*)"
    latency_pattern = r"[L,l]atency:\s*([0-9]*\.?[0-9]+)\s*([a-zA-Z/]*)"

    latency_list = []
    throughput_list = []
    latency_unit_name = ""
    throughput_unit_name = ""
    for idx, logfile in logfile_dict.items():
        with open(logfile, "r") as f:
            for line in f:
                re_latency = re.search(latency_pattern, line)
                re_throughput = re.search(throughput_pattern, line)
                if re_latency:
                    latency_list.append(float(re_latency.group(1)))
                    if not latency_unit_name:
                        latency_unit_name = re_latency.group(2)
                if re_throughput:
                    throughput_list.append(float(re_throughput.group(1)))
                    if not throughput_unit_name:
                        throughput_unit_name = re_throughput.group(2)
    if throughput_list and latency_list:
        assert (
            len(latency_list) == len(throughput_list) == len(logfile_dict)
        ), "Multiple instance benchmark failed with some instances!"

        # dump collected latency and throughput info
        header = "Multiple Instance Benchmark Summary"
        field_names = [
            "Instance",
            "Latency ({})".format(latency_unit_name),
            "Throughput ({})".format(throughput_unit_name),
        ]
        output_data = []
        for idx, (latency, throughput) in enumerate(zip(latency_list, throughput_list)):
            output_data.append([idx + 1, round(latency, 3), round(throughput, 3)])
        Statistics(output_data, header=header, field_names=field_names).print_stat()
        # show summary info
        logger.info("Average latency: {} {}".format(round(sum(latency_list) / len(latency_list), 3), latency_unit_name))
        logger.info("Total throughput: {} {}".format(round(sum(throughput_list), 3), throughput_unit_name))
    elif throughput_list:
        assert len(throughput_list) == len(logfile_dict), "Multiple instance benchmark failed with some instances!"

        # dump collected throughput info
        header = "Multiple Instance Benchmark Summary"
        field_names = [
            "Instance",
            "Throughput ({})".format(throughput_unit_name),
        ]
        output_data = []
        for idx, throughput in enumerate(throughput_list):
            output_data.append([idx + 1, round(throughput, 3)])
        Statistics(output_data, header=header, field_names=field_names).print_stat()
        # show summary info
        logger.info("Total throughput: {} {}.hdfghdfghs".format(round(sum(throughput_list), 3), throughput_unit_name))
    elif latency_list:
        assert len(latency_list) == len(logfile_dict), "Multiple instance benchmark failed with some instances!"

        # dump collected latency info
        header = "Multiple Instance Benchmark Summary"
        field_names = [
            "Instance",
            "Latency ({})".format(latency_unit_name),
        ]
        output_data = []
        for idx, latency in enumerate(latency_list):
            output_data.append([idx + 1, round(latency, 3)])
        Statistics(output_data, header=header, field_names=field_names).print_stat()
        # show summary info
        logger.info("Average latency: {} {}".format(round(sum(latency_list) / len(latency_list), 3), latency_unit_name))


def benchmark():
    """Benchmark API interface."""
    logger.info("Start benchmark with Intel Neural Compressor.")
    logger.info("Intel Neural Compressor only uses physical CPUs for the best performance.")

    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("--num_instances", type=int, default=None, help="Determine the number of instances.")
    parser.add_argument(
        "--num_cores_per_instance",
        type=int,
        default=None,
        help="Determine the number of cores in 1 instance.",
    )
    parser.add_argument("-C", "--cores", type=str, default=None, help="Determine the visible core range.")
    parser.add_argument("--cross_memory", action="store_true", help="Determine the visible core range.")
    parser.add_argument("script", type=str, help="The path to the script to launch.")
    parser.add_argument("parameters", nargs=argparse.REMAINDER, help="arguments to the script.")

    args = parser.parse_args()

    assert sys.platform in ["linux", "win32"], "only support platform windows and linux..."

    numa_info = dump_numa_info()  # show numa info and current usage of cores
    core_list_per_instance = set_cores_for_instance(args, numa_info=numa_info)
    script_and_parameters = args.script + " " + " ".join(args.parameters)
    logfile_dict = run_multi_instance_command(args, core_list_per_instance, raw_cmd=script_and_parameters)
    summary_latency_throughput(logfile_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CSE Graph Rewriter."""

from tensorflow.core.framework import graph_pb2

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class GraphCseOptimizer(GraphRewriterBase):
    """We introduce the CSE optimizer to optimize the nodes that contains identical op type.

    Case 1. Node A has three output nodes(B,C,D) and those child nodes has their own outputs
            (B1/C1C2/D1).
            Node A
            x    x    x
            x    x     x
    NODE B   NODE C  NODE D
    x        x   x     x
    B1       C1  C2    D1
    If Node B/C/D have the identical memory-bound op, like relu/relu6. The graph will be
    converted as below.
    We removed the Node C & Node D, updated the B as the input of C1/C2/D1.
            Node A
            x
            Node B
            x x  x  x
            x  x   x  x
        x   x    x  x
        B1  C1   C2  D1
    Case 2.  Node A has three output nodes(B,C,D) and those child nodes has their own outputs
            (B1/C1C2/D1).
            Node A
            x   x    x
            x    x     x
    NODE B   NODE C  NODE D
        x      x   x     x
        B1     C1  C2    D1
    If Node B and C have the identical memory-bound op, like relu/relu6. The graph will be
    converted as below.
    We removed the Node C, updated the B as the input of C1/C2.
                Node A
                x     x
            Node B  Node D
            x   x  x    x
            x   |   x    x
            x   |    x    x
            B1   C1    C2   D1

    Returns:
        [graphdef]: A optimized graphdef object.
    """

    computational_op_type = ("Conv2D", "Conv3D", "DepthwiseConv2dNative", "MatMul")

    @dump_elapsed_time("Pass GraphCseOptimizer")
    def do_transformation(self):
        """Optimize the graph contains multi output nodes.

        If those nodes' type are identical, those nodes should be elimated.
        Currently, we supported memory bound ops only.

        Args:
            input_graph_def (graphdef): graphdef object

        Returns:
            [graphdef]: optimized graph
        """
        GraphAnalyzer().graph = self.model

        graph_info = GraphAnalyzer().parse_graph()

        need_to_update_node = []
        # TODO Enhance below code snippet by using recursive method.
        for _, i in graph_info.items():
            candidate_node = [
                graph_info[child_name].node
                for child_name in i.outputs
                if graph_info[child_name].node.op not in self.computational_op_type
            ]
            candidate_node_unique_type = set([i.op for i in candidate_node])
            if len(candidate_node_unique_type) == len(candidate_node):
                # it means each sub node has their own type.
                continue
            node_type_name_mapping = {}
            # Created dict which key is op type and value is node has identical op type.
            for each_node in candidate_node:
                node_type = each_node.op
                node_name = each_node.name
                if node_type not in node_type_name_mapping:
                    node_type_name_mapping[node_type] = [node_name]
                else:
                    node_type_name_mapping[node_type].append(node_name)

            for _, node_names in node_type_name_mapping.items():
                # ignore unique op type and node with multi-outputs
                if len(node_names) == 1 or len(graph_info[node_names[0]].outputs) > 1:
                    continue
                # TODO Need to enhance below algorithm before golden.
                filter_node = [node_names[0]]
                for sub_node_name in node_names[1:]:
                    if not Helper.compare_node_attr(graph_info[node_names[0]].node, graph_info[sub_node_name].node):
                        continue
                    filter_node.append(sub_node_name)

                need_to_update_node.append({i.node.name: filter_node})

        for node_pair in need_to_update_node:
            for upper_node_name, lower_node_name in node_pair.items():
                keep_sub_node_name = lower_node_name[0]
                for removeable_node_name in lower_node_name[1:]:
                    graph_info[upper_node_name].outputs.remove(removeable_node_name)
                    for grand_child_node_name in graph_info[removeable_node_name].outputs:
                        filter_input_name = [
                            Helper.node_name_from_input(i) for i in graph_info[grand_child_node_name].node.input
                        ]
                        replace_index = filter_input_name.index(removeable_node_name)
                        graph_info[grand_child_node_name].node.input[replace_index] = keep_sub_node_name
                        graph_info[grand_child_node_name].node.input[replace_index] = keep_sub_node_name
                    graph_info.pop(removeable_node_name)

        output_graph_def = graph_pb2.GraphDef()

        for _, node_info in graph_info.items():
            output_graph_def.node.extend([node_info.node])

        return output_graph_def



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The utils for Tensorflow."""

from neural_compressor.tensorflow.utils.data import BaseDataLoader, DummyDataset, DummyDatasetV2
from neural_compressor.tensorflow.utils.model import (
    Model,
    TFConfig,
)
from neural_compressor.tensorflow.utils.constants import (
    SPR_BASE_VERSIONS,
    TENSORFLOW_DEFAULT_CONFIG,
    DEFAULT_SQ_ALPHA_ARGS,
    UNIFY_OP_TYPE_MAPPING,
)
from neural_compressor.tensorflow.utils.model_wrappers import (
    get_tf_model_type,
    BaseModel,
    KerasModel,
    TensorflowLLMModel,
    TensorflowBaseModel,
    TensorflowSavedModelModel,
)
from neural_compressor.tensorflow.utils.utility import (
    disable_random,
    algos_mapping,
    version1_lt_version2,
    version1_gt_version2,
    version1_eq_version2,
    version1_gte_version2,
    version1_lte_version2,
    register_algo,
    deep_get,
    itex_installed,
    dump_elapsed_time,
    combine_histogram,
    get_all_fp32_data,
    get_tensor_histogram,
    singleton,
    CpuInfo,
    Statistics,
    CaptureOutputToFile,
    valid_keras_format,
    TFSlimNetsFactory,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow Grappler Graph Rewriter."""

import tensorflow as tf
from tensorflow.core.protobuf import config_pb2, meta_graph_pb2
from tensorflow.python.grappler import tf_optimizer
from tensorflow.python.training import saver

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.utils import dump_elapsed_time, version1_gt_version2


class GrapplerOptimizer(GraphRewriterBase):
    """A python wrapper that leverages the built-in tensorflow grappler API to optimize the graph."""

    def __init__(self, model, input_output_names, opt_cfg):
        """Initialization."""
        super().__init__(model)
        self.input_output_names = input_output_names
        self.opt_cfg = opt_cfg
        self.generic_optimizer = ("pruning", "shape", "dependency", "debug_stripper", "loop")
        self.tf_2_optimizer = ("constfold", "arithmetic", "min_graph_nodes")

    @dump_elapsed_time("Pass GrapplerOptimizer")
    def do_transformation(self):
        """Apply tensorflow Grappler optimization."""
        try:
            g = tf.Graph()
            with g.as_default():
                g = tf.compat.v1.import_graph_def(self.model, name="")
                meta_graph = saver.export_meta_graph(graph_def=self.model, graph=g, clear_devices=True)
                fetch_collection = meta_graph_pb2.CollectionDef()
                for fetch in self.input_output_names:
                    fetch_collection.node_list.value.append(fetch)
                meta_graph.collection_def["train_op"].CopyFrom(fetch_collection)
                config = config_pb2.ConfigProto()
                rewriter_config = config.graph_options.rewrite_options
                for optimizer in self.generic_optimizer:
                    if optimizer in self.opt_cfg and self.opt_cfg[optimizer]:
                        rewriter_config.optimizers.append(optimizer)

                if version1_gt_version2(tf.version.VERSION, "2.2.0"):
                    for optimizer in self.tf_2_optimizer:
                        if optimizer in self.opt_cfg and self.opt_cfg[optimizer]:
                            rewriter_config.optimizers.append(optimizer)

                rewriter_config.min_graph_nodes = -1

                optimized_graph = tf_optimizer.OptimizeGraph(config, meta_graph)

            return optimized_graph
        except Exception as e:
            self.logger.warning("Fail to run grappler pass due to {}.".format(str(e)))
            return self.model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Remove training nodes Graph Rewriter."""

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class RemoveTrainingNodesOptimizer(GraphRewriterBase):
    """Remove training nodes optimizer."""

    def __init__(self, model, protected_nodes=[], types_to_splice=["Identity", "CheckNumerics", "StopGradient"]):
        """Initilizaiton."""
        super().__init__(model)
        self.protected_nodes = protected_nodes
        self.types_to_splice = types_to_splice

    @dump_elapsed_time("Pass RemoveTrainingNodesOptimizer")
    def do_transformation(self):
        """Remove tranining nodes which has no control edge inputs."""
        graph_handle = GraphAnalyzer()
        graph_handle.graph = self.model

        graph_info = graph_handle.parse_graph()
        # input_nodes = input_graph.node

        control_input_names = set()
        node_names_with_control_input = set()
        names_to_splice = {}

        for node_name, v in graph_info.items():
            for node_input in v.node.input:
                if "^" in node_input:
                    control_input_names.add(node_input.replace("^", ""))
                    node_names_with_control_input.add(node_name)

        for node_name, v in graph_info.items():
            if v.node.op in self.types_to_splice and v.node.name not in self.protected_nodes:
                # We don't want to remove nodes that have control edge inputs, because
                # they might be involved in subtle dependency issues that removing them
                # will jeopardize.
                if node_name not in node_names_with_control_input:
                    names_to_splice[node_name] = v.node.input[0]

        # We also don't want to remove nodes which are used as control edge inputs.
        names_to_splice = {name: value for name, value in names_to_splice.items() if name not in control_input_names}
        for k, _ in names_to_splice.items():
            graph_handle.remove_node_with_single_input_output(k)

        return graph_handle.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert LeakyRelu Graph Rewriter."""

from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class ConvertLeakyReluOptimizer(GraphRewriterBase):
    """Convert below subgraph to Node A + LeakyRelu.

    Node A             Node A
        |   x               |
        |    x              |
        |     Mul  --->     |
        |     x             |
        |   x               |
        Maximum           LeakyRelu
    Note, the coefficient of Mul should be less than 1 or the conversion is not valid.
    """

    @dump_elapsed_time("Pass ConvertLeakyReluOptimizer")
    def do_transformation(self):
        """Fuse small ops to LeakyRelu."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()
        target_nodes = g.query_fusion_pattern_nodes([["Mul"], ["Maximum"]])
        for i in target_nodes:
            successor_node_names = graph_info[i[1]].outputs

            mul_input_names = list(graph_info[i[0]].node.input)
            max_input_names = list(graph_info[i[1]].node.input)
            common_input = list(set(mul_input_names).intersection(set(max_input_names)))

            if len(common_input) != 1:
                continue
            mul_coeff_node_name = list(set(mul_input_names).difference(set(max_input_names)))[0]
            mul_coeff_node = graph_info[mul_coeff_node_name].node
            if mul_coeff_node.op != "Const":
                continue
            nd = tensor_util.MakeNdarray(mul_coeff_node.attr["value"].tensor).ndim
            if nd > 1:
                continue
            alpha_value = float(tensor_util.MakeNdarray(mul_coeff_node.attr["value"].tensor))
            if alpha_value > 1.0:
                continue

            leaky_relu_node_name = i[1] + "_leakyrelu"
            leaky_relu_node = Helper.create_node("LeakyRelu", leaky_relu_node_name, common_input)
            Helper.set_attr_dtype(leaky_relu_node, "T", dtypes.float32)
            Helper.set_attr_float(leaky_relu_node, "alpha", alpha_value)

            g.replace_single_node(leaky_relu_node, common_input, i[1], successor_node_names, i[1])
            g.remove_node(i[1])
            g.remove_node(i[0])
            g.remove_node(mul_coeff_node_name)

        return g.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initialize layer initializer functions."""

from neural_compressor.tensorflow.keras.layers.conv2d import initialize_int8_conv2d
from neural_compressor.tensorflow.keras.layers.dense import initialize_int8_dense
from neural_compressor.tensorflow.keras.layers.depthwise_conv2d import initialize_int8_depthwise_conv2d
from neural_compressor.tensorflow.keras.layers.pool2d import initialize_int8_avgpool, initialize_int8_maxpool
from neural_compressor.tensorflow.keras.layers.separable_conv2d import initialize_int8_separable_conv2d

layer_initializer_dict = {
    "QAvgPool2D": initialize_int8_avgpool,
    "QAveragePooling2D": initialize_int8_avgpool,
    "QMaxPool2D": initialize_int8_maxpool,
    "QMaxPooling2D": initialize_int8_maxpool,
    "QSeparableConv2D": initialize_int8_separable_conv2d,
    "QDepthwiseConv2D": initialize_int8_depthwise_conv2d,
    "QConv2D": initialize_int8_conv2d,
    "QDense": initialize_int8_dense,
}



import os
import re
import subprocess
import sys
from io import open

from setuptools import find_packages, setup


def fetch_requirements(path):
    with open(path, "r") as fd:
        return [r.strip() for r in fd.readlines()]


def is_commit_on_tag():
    try:
        result = subprocess.run(
            ["git", "describe", "--exact-match", "--tags"], capture_output=True, text=True, check=True
        )
        tag_name = result.stdout.strip()
        return tag_name
    except subprocess.CalledProcessError:
        return False


def get_build_version():
    if is_commit_on_tag():
        return __version__
    try:
        result = subprocess.run(["git", "describe", "--tags"], capture_output=True, text=True, check=True)
        distance = result.stdout.strip().split("-")[-2]
        commit = result.stdout.strip().split("-")[-1]
        return f"{__version__}.dev{distance}+{commit}"
    except subprocess.CalledProcessError:
        return __version__


try:
    filepath = "./neural_compressor/version.py"
    with open(filepath) as version_file:
        (__version__,) = re.findall('__version__ = "(.*)"', version_file.read())
except Exception as error:
    assert False, "Error: Could not open '%s' due %s\n" % (filepath, error)

PKG_INSTALL_CFG = {
    # overall installation config, pip install neural-compressor
    "neural_compressor": {
        "project_name": "neural_compressor",
        "include_packages": find_packages(
            include=["neural_compressor", "neural_compressor.*"],
            exclude=[
                "neural_compressor.template",
            ],
        ),
        "package_data": {"": ["*.yaml"]},
        "install_requires": fetch_requirements("requirements.txt"),
        "extras_require": {
            "pt": fetch_requirements("requirements_pt.txt"),
            "tf": fetch_requirements("requirements_tf.txt"),
        },
    },
    # 3.x pt binary build config, pip install neural-compressor-pt, install 3.x PyTorch API.
    "neural_compressor_pt": {
        "project_name": "neural_compressor_pt",
        "include_packages": find_packages(
            include=[
                "neural_compressor.common",
                "neural_compressor.common.*",
                "neural_compressor.torch",
                "neural_compressor.torch.*",
                "neural_compressor.transformers",
                "neural_compressor.transformers.*",
                "neural_compressor.evaluation",
                "neural_compressor.evaluation.*",
            ],
        ),
        "package_data": {"": ["*.json"]},
        "install_requires": fetch_requirements("requirements_pt.txt"),
    },
    # 3.x tf binary build config, pip install neural-compressor-tf, install 3.x TensorFlow API.
    "neural_compressor_tf": {
        "project_name": "neural_compressor_tf",
        "include_packages": find_packages(
            include=[
                "neural_compressor.common",
                "neural_compressor.common.*",
                "neural_compressor.tensorflow",
                "neural_compressor.tensorflow.*",
            ],
        ),
        "package_data": {"": ["*.yaml"]},
        "install_requires": fetch_requirements("requirements_tf.txt"),
    },
}


if __name__ == "__main__":
    cfg_key = "neural_compressor"

    # Temporary implementation of fp8 tensor saving and loading
    # Will remove after Habana torch applies below patch:
    # https://github.com/pytorch/pytorch/pull/114662
    ext_modules = []
    cmdclass = {}

    if "pt" in sys.argv:
        sys.argv.remove("pt")
        cfg_key = "neural_compressor_pt"

    if "tf" in sys.argv:
        sys.argv.remove("tf")
        cfg_key = "neural_compressor_tf"

    if bool(os.getenv("USE_FP8_CONVERT", False)):
        from torch.utils.cpp_extension import BuildExtension, CppExtension

        ext_modules = [
            CppExtension(
                "fp8_convert",
                ["neural_compressor/torch/algorithms/habana_fp8/tensor/convert.cpp"],
            ),
        ]
        cmdclass = {"build_ext": BuildExtension}

    project_name = PKG_INSTALL_CFG[cfg_key].get("project_name")
    include_packages = PKG_INSTALL_CFG[cfg_key].get("include_packages") or {}
    package_data = PKG_INSTALL_CFG[cfg_key].get("package_data") or {}
    install_requires = PKG_INSTALL_CFG[cfg_key].get("install_requires") or []
    extras_require = PKG_INSTALL_CFG[cfg_key].get("extras_require") or {}
    entry_points = {
        "console_scripts": [
            "incbench = neural_compressor.common.benchmark:benchmark",
        ]
    }

    setup(
        name=project_name,
        author="Intel AIPT Team",
        version=get_build_version(),
        author_email="feng.tian@intel.com, haihao.shen@intel.com, suyue.chen@intel.com",
        description="Repository of Intel® Neural Compressor",
        long_description=open("README.md", "r", encoding="utf-8").read(),
        long_description_content_type="text/markdown",
        keywords="quantization,auto-tuning,post-training static quantization,"
        "post-training dynamic quantization,quantization-aware training",
        license="Apache 2.0",
        url="https://github.com/intel/neural-compressor",
        packages=include_packages,
        include_package_data=True,
        package_data=package_data,
        install_requires=install_requires,
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        entry_points=entry_points,
        extras_require=extras_require,
        python_requires=">=3.7.0",
        classifiers=[
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "License :: OSI Approved :: Apache Software License",
        ],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Insert print node Graph Rewriter."""

import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_util as tu

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import version1_gt_version2


class InsertPrintMinMaxNode(GraphRewriterBase):
    """InsertPrintMinMaxNode Pass for tensorflow sampling."""

    def __init__(self, model, pre_node_name, post_node_name, new_api):
        """Initialization."""
        super().__init__(model)
        self.pre_node_name = pre_node_name
        self.post_node_name = post_node_name
        self.signature = pre_node_name + post_node_name
        self.new_api = new_api

    def do_transformation(self):
        """Insert print node in the graph to do the calibration."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        insert_node_pairs = []
        top_node = graph_info[self.pre_node_name].node
        if top_node.op == "ConcatV2":
            for i in range(top_node.attr["N"].i):
                insert_node_pairs.append([top_node.input[i], self.post_node_name])
        elif top_node.op in ("BatchMatMul", "BatchMatMulV2"):
            insert_node_pairs.append([top_node.input[0], self.post_node_name])
            if graph_info[top_node.input[1]].node.op != "Const":
                insert_node_pairs.append([top_node.input[1], self.post_node_name])
        elif top_node.op in ("Conv2DBackpropInput", "Conv3DBackpropInputV2"):
            insert_node_pairs.append([top_node.input[2], self.post_node_name])
        else:
            refresh_pre_node_name = graph_info[self.pre_node_name].node.input[0]
            # Check the Conv2D could be fused with previous Pad or not.
            # If so, we need to update the pre-node name correspondingly.
            refresh_pre_node = graph_info[Helper.node_name_from_input(refresh_pre_node_name)].node
            if refresh_pre_node.op == "Pad" and top_node.op in ("Conv2D", "Conv3D"):
                pad_const_node_name = refresh_pre_node.input[1]
                pad_const_node = graph_info[pad_const_node_name].node
                padding_tensor = None
                if graph_info[pad_const_node_name].node.op != "Const":
                    if pad_const_node.op == "DataFormatVecPermute":
                        parent_input_node = graph_info[pad_const_node.input[0]].node
                        if parent_input_node.op == "Const":
                            padding_tensor = tu.MakeNdarray(parent_input_node.attr["value"].tensor).flatten()
                else:
                    padding_tensor = tu.MakeNdarray(pad_const_node.attr["value"].tensor).flatten()
                if not any(padding_tensor) or (
                    any(padding_tensor) and (tf.version.VERSION == "1.15.0-up3" or self.new_api)
                ):
                    insert_node_pairs.append([refresh_pre_node_name, self.post_node_name])
                    refresh_pre_node_name = refresh_pre_node.input[0]

            insert_node_pairs.append([refresh_pre_node_name, self.post_node_name])

        output_names = []
        for node_pair_names in insert_node_pairs:
            for index, each_node_name in enumerate(node_pair_names):
                name_with_sig = each_node_name + self.signature
                node_name_prefix = name_with_sig.replace(":", "__port__").replace("^", "__hat__")
                reshape_dims_name = node_name_prefix + "_reshape_dims"
                reduction_dims_name = node_name_prefix + "_reduction_dims"

                reshape_dims_node = Helper.create_constant_node(reshape_dims_name, -1, dtypes.int32, [1])

                reduction_dims_node = Helper.create_constant_node(reduction_dims_name, 0, dtypes.int32, [1])

                # the training input QueueDequeueManyV2 has issue with implicit dependency
                # skip the input node of show_and_tell model
                if not (
                    Helper.node_name_from_input(each_node_name) == "batch_and_pad"
                    and graph_info[Helper.node_name_from_input(each_node_name)].node.op == "QueueDequeueManyV2"
                ):
                    reshape_dims_node.input.append("^" + Helper.node_name_from_input(each_node_name))
                    reduction_dims_node.input.append("^" + Helper.node_name_from_input(each_node_name))

                reshape_input_name = node_name_prefix + "_reshape_"

                reshape_input_node = Helper.create_node(
                    "Reshape", reshape_input_name, [each_node_name, reshape_dims_name]
                )

                min_input_name = node_name_prefix + "_min"
                min_input_node = Helper.create_node("Min", min_input_name, [reshape_input_name, reduction_dims_name])
                Helper.set_attr_dtype(min_input_node, "Tidx", dtypes.int32)
                Helper.set_attr_bool(min_input_node, "keep_dims", False)

                max_input_name = node_name_prefix + "_max"
                max_input_node = Helper.create_node("Max", max_input_name, [reshape_input_name, reduction_dims_name])
                Helper.set_attr_dtype(max_input_node, "Tidx", dtypes.int32)
                Helper.set_attr_bool(max_input_node, "keep_dims", False)

                max_print_node = Helper.create_node(
                    "Print",
                    node_name_prefix + "_print_max__{}".format(index),
                    [max_input_name + ":0", max_input_name + ":0"],
                )
                min_print_node = Helper.create_node(
                    "Print",
                    node_name_prefix + "_print_min__{}".format(index),
                    [min_input_name + ":0", min_input_name + ":0"],
                )

                if index == 0:
                    max_msg = ";{}_eightbit_max_{}__print__;__max:".format(self.pre_node_name, each_node_name)
                    min_msg = ";{}_eightbit_min_{}__print__;__min:".format(self.pre_node_name, each_node_name)
                    # workaround for swish_f32, attribute T is not in the op definition
                    if "swish_f32" in graph_info[self.pre_node_name].node.name:
                        src_dt = attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum)
                    else:
                        src_dt = graph_info[self.pre_node_name].node.attr["T"]
                else:
                    max_msg = ";{}_eightbit_requant_range__print__;__requant_max:".format(self.pre_node_name)
                    min_msg = ";{}_eightbit_requant_range__print__;__requant_min:".format(self.pre_node_name)
                    # workaround for swish_f32, attribute T is not in the op definition
                    if "swish_f32" in graph_info[each_node_name].node.op:
                        src_dt = attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum)
                    else:
                        src_dt = graph_info[each_node_name].node.attr["T"]

                reshape_input_node.attr["T"].CopyFrom(src_dt)
                min_input_node.attr["T"].CopyFrom(src_dt)
                min_print_node.attr["T"].CopyFrom(src_dt)
                max_input_node.attr["T"].CopyFrom(src_dt)
                max_print_node.attr["T"].CopyFrom(src_dt)

                min_print_node.attr["message"].s = min_msg.encode()
                min_print_node.attr["first_n"].i = -1
                min_print_node.attr["summarize"].i = 1024

                max_print_node.attr["message"].s = max_msg.encode()
                max_print_node.attr["first_n"].i = -1
                max_print_node.attr["summarize"].i = 1024

                attr_u = [dtypes.as_dtype(src_dt.type).as_datatype_enum]
                min_print_node.attr["U"].list.CopyFrom(attr_value_pb2.AttrValue.ListValue(type=attr_u))
                max_print_node.attr["U"].list.CopyFrom(attr_value_pb2.AttrValue.ListValue(type=attr_u))
                post_node_names = graph_info[Helper.node_name_from_input(each_node_name)].outputs
                if post_node_names:
                    for post_node_name in post_node_names:
                        post_node = graph_info[post_node_name].node
                        if each_node_name not in post_node.input:
                            continue
                        if (
                            post_node.op == "FusedBatchNormV3"
                            and "_print_identity"
                            not in graph_info[Helper.node_name_from_input(post_node.name)].node.input[0]
                        ):
                            identity_node = Helper.create_node(
                                "Identity",
                                post_node.name + "_print_identity",
                                [graph_info[Helper.node_name_from_input(post_node.name)].node.input[0]],
                            )
                            identity_node.attr["T"].CopyFrom(src_dt)
                            cur_graph.add_node(
                                identity_node,
                                graph_info[Helper.node_name_from_input(post_node.name)].node.input[0],
                                [post_node.name],
                            )
                            identity_node.input.append("^" + min_print_node.name)
                            identity_node.input.append("^" + max_print_node.name)
                        else:
                            post_node.input.append("^" + min_print_node.name)
                            post_node.input.append("^" + max_print_node.name)

                    cur_graph.add_node(reshape_dims_node, None, [reshape_input_name])
                    cur_graph.add_node(reduction_dims_node, None, [max_input_name, min_input_name])
                    cur_graph.add_node(reshape_input_node, each_node_name, [max_input_name, min_input_name])
                    cur_graph.add_node(max_input_node, reshape_input_name, [max_print_node.name])
                    cur_graph.add_node(min_input_node, reshape_input_name, [min_print_node.name])

                    cur_graph.add_node(min_print_node, min_input_name, [])
                    cur_graph.add_node(max_print_node, max_input_name, [])
                else:
                    identity_node0 = Helper.create_node(
                        "Identity", min_print_node.name + "_identity", [min_print_node.name]
                    )
                    identity_node0.attr["T"].CopyFrom(src_dt)
                    identity_node1 = Helper.create_node(
                        "Identity", max_print_node.name + "_identity", [max_print_node.name]
                    )
                    identity_node1.attr["T"].CopyFrom(src_dt)

                    cur_graph.add_node(reshape_dims_node, None, [reshape_input_name])
                    cur_graph.add_node(reduction_dims_node, None, [max_input_name, min_input_name])
                    cur_graph.add_node(reshape_input_node, each_node_name, [max_input_name, min_input_name])
                    cur_graph.add_node(max_input_node, reshape_input_name, [max_print_node.name])
                    cur_graph.add_node(min_input_node, reshape_input_name, [min_print_node.name])
                    cur_graph.add_node(min_print_node, min_input_name, [identity_node0.name])
                    cur_graph.add_node(max_print_node, max_input_name, [identity_node1.name])
                    cur_graph.add_node(identity_node0, min_print_node.name, [])
                    cur_graph.add_node(identity_node1, max_print_node.name, [])
                    # identity_node0.input.append("^" + min_print_node.name)
                    # identity_node1.input.append("^" + max_print_node.name)
                    output_names.append(identity_node0.name)
                    output_names.append(identity_node1.name)
        return cur_graph.dump_graph(), output_names



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""GraphTransform Base Class."""

from __future__ import absolute_import, division, print_function

import logging
import re

from tensorflow.core.framework import graph_pb2
from tensorflow.python.platform import gfile

logger = logging.getLogger("neural_compressor")


class GraphTransformBase(object):
    """GraphTransform Base Class."""

    def __init__(self, input_pb):
        """Basic class for graph transformation.

        Args:
             input_pb: the input graphdef or pb file.
        """
        if isinstance(input_pb, graph_pb2.GraphDef):
            self.input_graph = input_pb
        else:
            try:
                with gfile.Open(input_pb, "rb") as f:
                    self.input_graph.ParseFromString(f.read())
            except Exception as e:
                logger.error("Fail to read input pb from {} due to {}.".format(input_pb, str(e)))

        self.node_mapping = {}
        self.node_name_list = []
        self.output_node_map = {}
        self.generate_input_map()

    def parse_input_pb(self):
        """Parse the input pbdef to get the node name and node mapping.

        Returns:
            the dict that key is node name while the value is nodeDef.
        """
        for node in self.input_graph.node:
            self.node_name_list.append(node.name)

            if node.name not in self.node_mapping:
                self.node_mapping[node.name] = node
            else:
                logger.warning("Duplicated node name {}.".format(node.name))

    def generate_input_map(self):
        """Generate the input map."""
        self.input_node_map = {}
        for node in self.input_graph.node:
            node_name = self.node_name_from_input(node.name)
            if node_name not in self.input_node_map:
                self.input_node_map[node_name] = node
            else:
                raise ValueError("Duplicate node names detected for ", node.name)

    def node_name_from_input(self, node_name):
        """Get the original node name from input string.

        Args:
            node_name: input node's name in string

        Returns:
            node's name
        """
        if node_name.startswith("^"):  # pragma: no cover
            node_name = node_name[1:]
        m = re.search(r"(.*):\d+$", node_name)
        if m:
            node_name = m.group(1)
        return node_name

    def get_node_name_from_input(self, node_name):
        """Get the original node name from input string.

        Args:
            node_name: input node's name in string

        Returns:
            node's name
        """
        node_names = node_name.split(":")
        return node_names[0]

    def do_transformation(self):
        """Virtual Interface.

        Each transformation should implement it.
        """
        pass
